//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop
#include <math.h>

#include "xhpgl.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TXHPGLBox *XHPGLBox;

void __fastcall TXHPGLBox::FormClose(TObject *Sender, TCloseAction &Action)
{
	STLIST *stnext;
	clear_pltlist();

	// clear steg-liste
	while(stlist)
	{
		stnext=stlist->next;
		free(stlist);
		stlist=stnext;
	}
}
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
__fastcall TXHPGLBox::TXHPGLBox(TComponent* Owner)
	: TForm(Owner)
{
	pltlist=NULL;
	pltlast=NULL;
	stlist=NULL;
}

boolean TXHPGLBox::split_line_list(int width, int x1, int y1, int x2, int y2, int *xa, int *ya, int *xb, int *yb)
{
	STLIST *stnext;

	stnext=stlist;
	while(stnext)
	{
		if(split_line(stnext->x, stnext->y, width, x1, y1, x2, y2, xa, ya, xb, yb))
		{
			return true;
		}
		stnext=stnext->next;
	}
	return false;
}

boolean TXHPGLBox::split_line(int x, int y, int width, int x1, int y1, int x2, int y2, int *xa, int *ya, int *xb, int *yb)
{
	int x3,y3;
	int px,py,px1,py1;
	double ax;

	x3=x;
	y3=y;

	// ist der angegebene punkt (x,y) auf der linie?
	if(find_lpoint_pos(x1, y1, x2, y2, 1, &x3, &y3))
	{
		if(x1==x2)
		{
			// keine nderung x-achse; nur auf der y-achse wird unterbrochen
			*xa=x1;
			*xb=x1;
			if(y1<y2)
			{
				*ya=y3-width/2;
				*yb=y3+width/2;
			}
			else
			{
				*ya=y3+width/2;
				*yb=y3-width/2;
			}
		}
		else if(y1==y2)
		{
			// keine nderung y-achse; nur auf der x-achse wird unterbrochen
			*ya=y1;
			*yb=y1;
			if(x1<x2)
			{
				*xa=x3-width/2;
				*xb=x3+width/2;
			}
			else
			{
				*xa=x3+width/2;
				*xb=x3-width/2;
			}
		}
		else
		{
			// der punkt ist auf der linie, errechne die linienunterbrechungen
			ax=atan2((x1-x2),(y1-y2));

			// dies ist der 2. punkt (x1,y1 ist der 1.; x2,y2 ist der 4.)
			*xa=x3+(double)(width/2)*sin(ax);
			*ya=y3+(double)(width/2)*cos(ax);
			// dies ist der 3. punkt (x1,y1 ist der 1.; x2,y2 ist der 4.)
			*xb=x3-(double)(width/2)*sin(ax);
			*yb=y3-(double)(width/2)*cos(ax);
		}
		return true;
	}
	return false;
}


boolean TXHPGLBox::find_lpoint_pos(int x1, int y1, int x2, int y2, int snap, int *x3, int *y3)
{
	int xx,xy;
	int px,py,px1,py1;
	int dx,dy;
	double vx;
	double ax;

	xx=*x3; xy=*y3;

	px=x1; px1=x2;
	py=y1; py1=y2;
	if(x1>x2)
	{
		px=x2;
		px1=x1;
	}
	if(y1>y2)
	{
		py=y2;
		py1=y1;
	}

	// ist diese linie im bereich der coordinaten?
	if(xx > px-snap && xx < px1+snap && xy > py-snap && xy < py1+snap)
	{
		dx=px-px1;
		dy=py-py1;

		if(dx==0 || dy==0)
		{
			// keine nderung in x oder y
			if(dy==0)
			{
				if(xx>px && xx <px1)
				{
					// keine nderung in y und x im bereich
					*x3=xx; *y3=py;
					return true;
				}
			}
			else
			{
				if(xy>py && xy <py1)
				{
					// keine nderung in x und y im bereich
					*x3=px; *y3=xy;
					return true;
				}
			}
		}
		else
		{
			vx=(double)(x1-x2)/(double)(y1-y2);
			// hier mu berechnet werden
			dx=((double)(xy-y1)*vx)+x1;
			dy=((double)(xx-x1)/vx)+y1;
			if(xx>dx-10 && xx < dx+10 && xy>dy-10 && xy<dy+10)
			{
				*x3=dx; *y3=xy;
				return true;
			}
		}
	}
	return false;
}

void TXHPGLBox::ST_add(int x, int y)
{
	STLIST *next;

	next = (STLIST *)malloc(sizeof(STLIST));
	if(next)
	{
		next->x=x;
		next->y=y;
		next->next=stlist;
		stlist=next;
	}
}

void TXHPGLBox::ST_rm(int x, int y)
{
	STLIST *stnext;
	STLIST *stprev;

	stnext = stlist;
	stprev=NULL;
	while(stnext)
	{
		if(x > stnext->x-40 && x < stnext->x+40 && y > stnext->y-40 && y < stnext->y+40)
		{
			if(stprev==NULL)
			{
				stlist=stlist->next;
			}
			else
			{
				stprev->next=stnext->next;
			}
			free(stnext);
			break;
		}
		stprev=stnext;
		stnext=stnext->next;
	}
}


void TXHPGLBox::clear_pltlist(void)
{
	PLTLIST *pnext;
	PLTLIST *ptmp;
	CORDLIST *cnext;
	CORDLIST *ctmp;


	pnext=pltlist;
	while(pnext)
	{
		ptmp=pnext->next;
		cnext=pnext->cord;
		while(cnext)
		{
			ctmp=cnext->next;
			free(cnext);
			cnext=ctmp;
		}
		free(pnext);
		pnext=ptmp;
	}

	pltlist=NULL;
	pltlast=NULL;
}

void TXHPGLBox::add_pltlist(void)
{
	PLTLIST *next;

	next = (PLTLIST *)malloc(sizeof(PLTLIST));
	if(next)
	{
		next->next=NULL;
		next->cord=NULL;
		next->cordl=NULL;
		next->rec_x=0;
		next->rec_y=0;
		next->rec_x1=0;
		next->rec_y1=0;
		next->del=0;
		next->mark=0;
		if(!pltlast)
		{
			pltlist=next;
		}
		else
		{
			pltlast->next=next;
		}
		pltlast=next;
	}
}

void TXHPGLBox::add_cord(int x, int y, int z, int a,  int sp, PLTLIST *cur)
{
	CORDLIST *next;
	if(cur)
	{
		next = (CORDLIST *)malloc(sizeof(CORDLIST));
		if(next)
		{
			next->next=NULL;
			next->prev=cur->cordl;
			next->x=x;
			next->y=y;
			next->z=z;
			next->a=a;
			next->sp=sp;


			// remember outlines
			if(x>cur->rec_x1)
			{
				cur->rec_x1=x;
			}
			if(x<cur->rec_x)
			{
				cur->rec_x=x;
			}
			if(y>cur->rec_y1)
			{
				cur->rec_y1=y;
			}
			if(y<cur->rec_y)
			{
				cur->rec_y=y;
			}

			if(!cur->cordl)
			{
				cur->cord=next;
			}
			else
			{
				cur->cordl->next=next;
			}

			// letzte coordinate in der liste merken
			cur->cordl=next;
		}
	}
}

void TXHPGLBox::cp_pltlist(TMDIChild *mic)
{
	PLTLIST *pnext;
	CORDLIST *cnext;
	STLIST *stnext;
	int x,y;

	pnext=mic->pltlist;
	while(pnext)
	{
		if(pnext->del == false)
		{
			add_pltlist();
			pltlast->cord=NULL;
			pltlast->cordl=NULL;
			pltlast->rec_x=pnext->rec_x;
			pltlast->rec_y=pnext->rec_y;
			pltlast->rec_x1=pnext->rec_x1;
			pltlast->rec_y1=pnext->rec_y1;
			pltlast->del=pnext->del;
			pltlast->mark=pnext->mark;

			cnext=pnext->cord;
			while(cnext)
			{
				x=cnext->x;
				y=cnext->y;
				mic->transform_save(&x,&y);
				add_cord(x, y,cnext->z,cnext->a,cnext->sp,pltlast);
				cnext=cnext->next;
			}
		}
		pnext=pnext->next;
	}

	// die liste der stege kopieren
	stnext=mic->stlist;
	while(stnext)
	{
		ST_add(stnext->x, stnext->y);
		stnext=stnext->next;
    }
}

void TXHPGLBox::wopt_pltlist(void)
{
	PLTLIST *pnext;
	PLTLIST *p1next;
	CORDLIST *c1next;
	boolean is_opt = false;

	pnext=pltlist;
	while(pnext)
	{
		// suche in der nachfolgende liste eine coordinate die gleich dem ende oder dem
		// anfang einer anderen coordinate ist

		if(pnext->del == false)
		{
			// ist eine linie vorhanden, werden die coordinaten angehngt und die quelle als gelscht gekennzeichnet
			p1next=pnext->next;
			is_opt=false;
			while(p1next)
			{
				// erst wird nur grob nachgesehen ob das umschlieende rechteck das gleiche ist
				// dies kann bei unterschiedlichen frslinien durchaus gleich sein
				// ist aber eine schnelle methode um nicht immer alle frskoordinaten prfen zu mssen
				if(p1next->del == false)
				{
					// pat das ende an einen anfang?
					if(pnext->cordl->x == p1next->cord->x && pnext->cordl->y == p1next->cord->y)
					{
						p1next->del=true;
						// zeige auf die nchste koordinate
						c1next=p1next->cord->next;
						while(c1next)
						{
							// fge die koordinate zu
							add_cord(c1next->x, c1next->y, c1next->z, c1next->a,  c1next->sp, pnext);
							c1next=c1next->next;
						}
						is_opt=true;
						break;
					}
					// pat das ende an ein ende?
					else if(pnext->cordl->x == p1next->cordl->x && pnext->cordl->y == p1next->cordl->y)
					{
						p1next->del=true;
						// zeige auf die nchste koordinate
						c1next=p1next->cordl->prev;
						while(c1next)
						{
							// fge die koordinate zu
							add_cord(c1next->x, c1next->y, c1next->z, c1next->a,  c1next->sp, pnext);
							c1next=c1next->prev;
						}
						is_opt=true;
						break;
					}
				}
				p1next=p1next->next;
			}
		}
		if(is_opt==true) continue;
		pnext=pnext->next;
	}
}

void TXHPGLBox::zs_pltlist(void)
{
	PLTLIST *pnext;
	PLTLIST *pl = pltlast;
	CORDLIST *cnext;
	boolean first_loop=true;

	int zustellung,max_tiefe,tiefe, v_draw;

	v_draw=atoi(Draw_v->Text.c_str());

	zustellung=(atof(Zs->Text.c_str())*40.0);
	max_tiefe=(atof(Down->Text.c_str())*40.0);
	tiefe=max_tiefe;
	if(CheckBox1->Checked)
	{
		tiefe=0;
	}

	do
	{
		if(CheckBox1->Checked)
		{
			tiefe+=zustellung;
			if(tiefe > max_tiefe)
			{
				tiefe=max_tiefe;
			}
		}

		pnext=pltlist;
		while(pnext)
		{
			if(pnext->del == false)
			{
				if(first_loop)
				{
					cnext=pnext->cord;
					while(cnext)
					{
						if(RedefPar->Checked || cnext->sp==0)
						{
							cnext->z=tiefe;
						}
						cnext=cnext->next;
					}
				}
				else
				{
					add_pltlist();
					pltlast->cord=NULL;
					pltlast->cordl=NULL;
					pltlast->rec_x=pnext->rec_x;
					pltlast->rec_y=pnext->rec_y;
					pltlast->rec_x1=pnext->rec_x1;
					pltlast->rec_y1=pnext->rec_y1;
					pltlast->del=pnext->del;
					pltlast->mark=pnext->mark;

					cnext=pnext->cord;
					while(cnext)
					{
						if(RedefPar->Checked || cnext->sp==0)
						{
							add_cord(cnext->x, cnext->y,tiefe,cnext->a,v_draw,pltlast);
						}
						else
						{
							add_cord(cnext->x, cnext->y,cnext->z,cnext->a,cnext->sp,pltlast);
						}
						cnext=cnext->next;
					}
				}
			}
			if(pnext==pl)
			{
				first_loop=false;
				break;
			}
			pnext=pnext->next;
		}
	} while(tiefe<max_tiefe);

	// geschwindigkeit setzen
	pnext=pltlist;
	while(pnext)
	{
		if(pnext->del == false)
		{
			cnext=pnext->cord;
			while(cnext)
			{
				if(RedefPar->Checked || cnext->sp==0)
				{
					cnext->sp=v_draw;
				}
				cnext=cnext->next;
			}
		}
		pnext=pnext->next;
	}
}

void TXHPGLBox::opt_pltlist(void)
{
	PLTLIST *pnext;
	PLTLIST *p1next;
	CORDLIST *cnext;
	CORDLIST *c1next;
	int to_del;

	pnext=pltlist;
	while(pnext)
	{
		// wenn nicht gelscht suche in der nachfolgende liste eine frsung mit den gleichen frslinien
		// ist eine linie vorhanden, wird diese gelscht (als gelscht gekennzeichnet)
		if(pnext->del == false)
		{
			p1next=pnext->next;
			while(p1next)
			{
				// erst wird nur grob nachgesehen ob das umschlieende rechteck das gleiche ist
				// dies kann bei unterschiedlichen frslinien durchaus gleich sein
				// ist aber eine schnelle methode um nicht immer alle frskoordinaten prfen zu mssen
				if(p1next->del == false && pnext->rec_x == p1next->rec_x && pnext->rec_y == p1next->rec_y && pnext->rec_x1 == p1next->rec_x1 && pnext->rec_y1 == p1next->rec_y1)
				{
					p1next->del=true;
					// das umschlieende rechteck ist gleich, nun werden ALLE coordinaten die zu
					// frsen sind verglichen
					cnext=pnext->cord;
					c1next=p1next->cord;
					while(cnext && c1next)
					{
						if(cnext->x != c1next->x || cnext->y != c1next->y || cnext->a != c1next->a || cnext->z < c1next->z)
						{
							// eine der koordinaten (oder die tiefe) sind anders.. nicht lschen!
							p1next->del=false;
							break;
						}
						cnext=cnext->next;
						c1next=c1next->next;
					}
					if(cnext == NULL && c1next != NULL)
					{
						// ist einer der zeiger NULL und der ander nicht, nicht lschen
						p1next->del=false;
					}
					if(c1next == NULL && cnext != NULL)
					{
						// ist einer der zeiger NULL und der ander nicht, nicht lschen
						p1next->del=false;
					}
				}
				p1next=p1next->next;
			}
		}
		pnext=pnext->next;
	}
}

void __fastcall TXHPGLBox::save_xplt_file(FILE *fp)
{
	char buf[2048];
	PLTLIST *pnext;
	CORDLIST *cnext;
	int x,y,a,len;
	int xa,ya,xb,yb;
	int lpen=-99;

	int old_x,old_y,old_a;
	int up_p, up_p_10, v_move, v_down, home;
	int st_width,st_deep;

	v_down=atoi(Down_v->Text.c_str());
	v_move=atoi(Move_v->Text.c_str());

	st_width=((atof(frb->Text.c_str())+atof(stb->Text.c_str()))*40.0);
	st_deep=(atof(stt->Text.c_str())*40.0);

	if(!CheckBox1->Checked)
	{
		// wenn keine zustellung ist die steghhe 0
		st_deep=0;
	}

	up_p=(atof(Up->Text.c_str())*40.0);
	up_p_10=(atof(Up->Text.c_str())*40.0)/10;

	home=(atof(Home->Text.c_str())*40.0);

	old_x=0;
	old_y=0;
	old_a=0;

	sprintf(buf,"XHPGL;\nPA 0,0,-%d,0,%d;\n",home,v_move);
	fputs(buf,fp);

	pnext=pltlist;
	while(pnext)
	{
		// nur wenn nicht gelscht ist
		if(pnext->del==false)
		{
			// nur wenn mindestens ein koordinatenpaar vorhanden ist
			if(pnext->cord && pnext->cord->next)
			{
				x=pnext->cord->x;
				y=pnext->cord->y;
				a=pnext->cord->a;

				// pen Up
				sprintf(buf,"PA %d,%d,-%d,%d,%d;\n",old_x,old_y,up_p,old_a,v_move);
				fputs(buf,fp);

				// pen move
				sprintf(buf,"PA %d,%d,-%d,%d,%d;\n",x,y,up_p,old_a,v_move);
				fputs(buf,fp);
				// fast pen down and slow pen down
				sprintf(buf,"PA %d,%d,-%d,%d,%d;\nPA %d,%d,%d,%d,%d;\n",x,y,up_p_10,old_a,v_move, x,y,pnext->cord->z,old_a,v_down);
				fputs(buf,fp);

				old_x=x;
				old_y=y;
				old_a=a;

				cnext=pnext->cord->next;
				while(cnext)
				{
					x=cnext->x;
					y=cnext->y;
					a=cnext->a;


					if(StOn->Checked && cnext->z > st_deep && split_line_list(st_width, old_x, old_y, x, y, &xa, &ya, &xb, &yb))
					{
						// stege bercksichtigen
						len=sprintf(buf,"PA %d,%d,%d,%d,%d;\n",xa,ya,cnext->z,cnext->a,cnext->sp);
						len+=sprintf(buf+len,"PA %d,%d,%d,%d,%d;\n",xa,ya,st_deep,cnext->a,cnext->sp);
						len+=sprintf(buf+len,"PA %d,%d,%d,%d,%d;\n",xb,yb,st_deep,cnext->a,cnext->sp);
						len+=sprintf(buf+len,"PA %d,%d,%d,%d,%d;\n",xb,yb,cnext->z,cnext->a,v_down);
						fputs(buf,fp);
					}
					sprintf(buf,"PA %d,%d,%d,%d,%d;\n",x,y,cnext->z,cnext->a,cnext->sp);
					fputs(buf,fp);
					old_x=x;
					old_y=y;
					old_a=a;
					cnext=cnext->next;
				}
			}
		}
		pnext=pnext->next;
	}

	// pen Up
	sprintf(buf,"PA %d,%d,-%d,%d,%d;\n",old_x,old_y,up_p,old_a,v_move);
	fputs(buf,fp);

	sprintf(buf,"PA 0,0,-%d,0,%d;\n",up_p,v_move);
	fputs(buf,fp);

	sprintf(buf,"PA 0,0,-%d,0,%d;\n",home,v_move);
	fputs(buf,fp);

}

//---------------------------------------------------------------------------
void __fastcall TXHPGLBox::Button2Click(TObject *Sender)
{
	int i;
	FILE *fp;
	AnsiString fn ="";
	TMDIChild *Child;
	char buf[1024];

	fn=ListBox1->Items->Strings[0];
	for(i=0;i<ListBox1->Count;i++)
	{
		if(ListBox1->Selected[i]==true)
		{
			fn=ListBox1->Items->Strings[i];
		}
	}
	if(SaveDialog1->FileName=="")
	{
		SaveDialog1->FileName=fn.Delete(fn.LastDelimiter("."),9999)+".xplt";
	}
	if (SaveDialog1->Execute())
	{
		// save XHPGL
		if((fp=fopen(SaveDialog1->FileName.c_str(),"w"))!=NULL)
		{
			for(i=0;i<ListBox1->Count;i++)
			{
				if(ListBox1->Selected[i]==true)
				{
					Child=(TMDIChild *)ListBox1->Items->Objects[i];
					cp_pltlist(Child);
					// daten speichern unter SaveDialog1->FileName.c_str()
				}
			}

			if(WegOpt->Checked)
			{
				wopt_pltlist();
			}

			zs_pltlist();

			if(OptOut->Checked)
			{
				// optimiere pltliste
				opt_pltlist();
			}

			// wird die wegeoptimierung verwendet, wird durch einen weiteren
			// aufruf der wegeoptimierung nach dem verwenden der zustellung
			// auch das heben und senken an dem startpunkt unterbunden
			// da die eintrge in der liste mit steigender tiefe auch weiter
			// hinten ist, kann diese optimierung bedenkenlos ausgefhrt werden

			if(WegOpt->Checked)
			{
				wopt_pltlist();
			}

			save_xplt_file(fp);
			fclose(fp);
			clear_pltlist();
		}
		ModalResult=mrOk;
	}
}

int lb_index=-1;
//---------------------------------------------------------------------------
void __fastcall TXHPGLBox::ListBox1DragDrop(TObject *Sender, TObject *Source,
	  int X, int Y)
{
	int idx;
	if((idx=ListBox1->ItemAtPos(TPoint(X,Y),true))>=0 && lb_index>=0)
	{
		ListBox1->Items->Move(lb_index,idx);
	}
}
//---------------------------------------------------------------------------
void __fastcall TXHPGLBox::ListBox1MouseDown(TObject *Sender,
	  TMouseButton Button, TShiftState Shift, int X, int Y)
{
	lb_index=ListBox1->ItemAtPos(TPoint(X,Y),true);
}
//---------------------------------------------------------------------------
void __fastcall TXHPGLBox::ListBox1DragOver(TObject *Sender, TObject *Source,
	  int X, int Y, TDragState State, bool &Accept)
{
	Accept = Source==ListBox1;
}
//---------------------------------------------------------------------------




